import pytest

from tokenizers.pre_tokenizers import PreTokenizer

from dom_tokenizers.internal.transformers import AutoTokenizer
from dom_tokenizers.pre_tokenizers import DOMSnapshotPreTokenizer
from dom_tokenizers.train import DEFAULT_BASE_TOKENIZER


@pytest.fixture
def base_tokenizer():
    """An instance of the default base tokenizer we train our
    tokenizers from.
    """
    return AutoTokenizer.from_pretrained(DEFAULT_BASE_TOKENIZER)


@pytest.fixture
def dom_snapshot_pre_tokenizer():
    """An instance of a pre-tokenizer that consumes JSON-serialized
    DOM snapshots.
    """
    return PreTokenizer.custom(DOMSnapshotPreTokenizer())



import json
import warnings

from argparse import ArgumentParser
from math import log10, floor

from datasets import load_dataset
from tokenizers.pre_tokenizers import PreTokenizer, WhitespaceSplit

from .internal.transformers import AutoTokenizer
from .pre_tokenizers import DOMSnapshotPreTokenizer

DEFAULT_BASE_TOKENIZER = "bert-base-cased"
DEFAULT_SPLIT = "train"
DEFAULT_SIZE = 1024
SEND_BUGS_TO = "https://github.com/gbenson/dom-tokenizers/issues"


def train_tokenizer(
        training_dataset,
        base_tokenizer=DEFAULT_BASE_TOKENIZER,
        vocab_size=DEFAULT_SIZE,
        corpus_size=None):

    # Create the base tokenizer we'll train our new tokenizer from.
    if isinstance(base_tokenizer, str):
        base_tokenizer = AutoTokenizer.from_pretrained(base_tokenizer)

    # Create the custom pretokenizer our new tokenizer will use.
    new_pretokenizer = DOMSnapshotPreTokenizer()

    # List the custom special tokens that need adding to our tokenizer.
    new_special_tokens = [
        special_token
        for special_token in new_pretokenizer.special_tokens
        if base_tokenizer.tokenize(special_token) != [special_token]
    ]

    # It's not possible to train using a custom pre-tokenizer, the Rust
    # code raises "Exception: Custom PreTokenizer cannot be serialized"
    # (see https://github.com/huggingface/tokenizers/issues/269) so we
    # have to run our pre-tokenizer manually, then join its output with
    # whitespace and hope the regular pretokenizer takes it back apart
    # how we need it to.

    base_tokenizer.backend_tokenizer.pre_tokenizer = WhitespaceSplit()
    base_pretokenizer = base_tokenizer.backend_tokenizer.pre_tokenizer
    new_pretokenizer = PreTokenizer.custom(new_pretokenizer)

    def futz_input(real_input):
        pretokenized = new_pretokenizer.pre_tokenize_str(real_input)
        want_tokens = [token for token, offsets in pretokenized]
        futzed_input = " ".join(want_tokens)
        pretokenized = base_pretokenizer.pre_tokenize_str(futzed_input)
        got_tokens = [token for token, offsets in pretokenized]
        assert got_tokens == want_tokens
        return futzed_input

    def get_training_corpus():
        for row in training_dataset:
            yield futz_input(json.dumps(row["dom_snapshot"]))

    # Try and get a dataset length, for the progress tracker.
    if corpus_size is None:
        try:
            corpus_size = len(training_dataset)
        except TypeError:
            pass
    cs = f"{corpus_size:,}" if corpus_size else "an unknown number of"
    print(f"Generating {vocab_size:,} tokens from {cs} examples:")

    # Train the new tokenizer.
    new_tokenizer = base_tokenizer.train_new_from_iterator(
        text_iterator=get_training_corpus(),
        vocab_size=vocab_size,
        new_special_tokens=new_special_tokens,
        length=corpus_size,
        show_progress=True,
    )

    # Post-training fixups.
    new_tokenizer.name_or_path = _pretty_name(new_tokenizer)
    for token in new_special_tokens:
        attr = f"{token[1:-1].lower()}_token"
        if not hasattr(new_tokenizer, attr):
            continue  # not a "special" special token
        if getattr(new_tokenizer, attr) is not None:
            continue  # already got this one
        setattr(new_tokenizer, attr, token)

    return new_tokenizer


def _pretty_name(tokenizer=None, *, vocab_size=None, prefix="dom-tokenizer-"):
    if vocab_size is None:
        vocab_size = tokenizer.vocab_size
    pretty_size = _round_and_prefix(vocab_size)
    return f"{prefix}{pretty_size}"


def _round_and_prefix(value):
    """314159 -> '314k'."""
    whole, frac = divmod(log10(value), 1)
    unit_index, whole = divmod(floor(whole), 3)
    value = round(10 ** (whole + frac))
    unit = ([""] + list("kMBTQ"))[unit_index]
    return f"{value}{unit}"


def main():
    p = ArgumentParser(
        description="Train DOM-aware tokenizers.",
        epilog=f"Report bugs to: <{SEND_BUGS_TO}>.")
    p.add_argument(
        "dataset", metavar="DATASET",
        help="dataset containing the training corpus")
    p.add_argument(
        "--base-tokenizer", metavar="ID",
        default=DEFAULT_BASE_TOKENIZER,
        help=(f"tokenizer to train ours from"
              f" [default: {DEFAULT_BASE_TOKENIZER}]"))
    p.add_argument(
        "--split", default=DEFAULT_SPLIT, metavar="SPLIT", dest="split_name",
        help=(f"split of the training dataset to use"
              f" [default: {DEFAULT_SPLIT}]"))
    p.add_argument(
        "-N", "--num-inputs", metavar="N", dest="corpus_size",
        type=int,
        help=("number of sequences in the training dataset, if known;"
              " this is used to provide meaningful progress tracking"))
    p.add_argument(
        "-n", "--num-tokens", metavar="N", dest="vocab_size", type=int,
        default=DEFAULT_SIZE,
        help=(f"desired vocabulary size, including all special tokens and"
              f" the initial alphabet [default: {DEFAULT_SIZE} tokens]"))
    p.add_argument(
        "-o", "--output", metavar="DIR", dest="save_directory",
        help=("directory to save the trained tokenizer into"
              " [default: something based on targeted vocabulary size]"))
    args = p.parse_args()

    save_directory = args.save_directory
    if save_directory is None:
        save_directory = _pretty_name(vocab_size=args.vocab_size)
        print(f"Output directory: {save_directory}\n")

    warnings.filterwarnings("ignore", message=r".*resume_download.*")

    tokenizer = train_tokenizer(
        load_dataset(
            args.dataset,
            split=args.split_name,
            streaming=True),
        base_tokenizer=args.base_tokenizer,
        vocab_size=args.vocab_size,
        corpus_size=args.corpus_size)
    print(f'\n{tokenizer.tokenize("training complete")}')

    tokenizer.save_pretrained(save_directory)

    print(tokenizer.tokenize("tokenizer state saved"))
    print(tokenizer.tokenize("see you soon") + ["!!"])



import os

# Don't print "None of PyTorch, TensorFlow >= 2.0, or Flax have been
# found. Models won't be available and only tokenizers, configuration
# and file/data utilities can be used" warning.  Tokenizers is all we
# want!

__var_name = "TRANSFORMERS_NO_ADVISORY_WARNINGS"
__orig_val = os.environ.get(__var_name)
os.environ[__var_name] = "1"
try:
    from transformers import AutoTokenizer  # noqa: F401
finally:
    if __orig_val is None:
        os.environ.pop(__var_name)
    else:  # pragma: no cover
        os.environ[__var_name] = __orig_val
    del __var_name, __orig_val, os



import os


def get_resource_filename(filename, *, ext=None):
    if ext and not filename.endswith(ext):
        filename += ext
    testdir = os.path.dirname(__file__)
    return os.path.join(testdir, "resources", filename)


def open_resource(filename, *, mode="r", encoding="utf-8", **kwargs):
    filename = get_resource_filename(filename, **kwargs)
    return open(filename, mode=mode, encoding=encoding)


def load_resource(filename, **kwargs):
    with open_resource(filename, **kwargs) as fp:
        return fp.read()



import json

from datasets import Dataset

from dom_tokenizers.train import train_tokenizer

from .util import load_resource


def test_base64(base_tokenizer, dom_snapshot_pre_tokenizer):
    """Test that base64 is entered successfully.  Incorrectly-sequenced
    lowercasing (i.e. applied prior to pre-tokenization) will cause this
    test to fail.
    """
    snapshot = load_resource("svg-in-base64.json")
    tokenizer = train_tokenizer(
        training_dataset=Dataset.from_dict({
            "dom_snapshot": (
                json.loads(snapshot),
            ),
        }),
    )
    tokenizer.backend_tokenizer.pre_tokenizer = dom_snapshot_pre_tokenizer

    tokens = tokenizer.tokenize(snapshot)
    print(tokens)

    # <img src="data:image/svg+xml;base64,PHN2Zy...2Zz4=" width="256"...
    expect = [
        "[TAG]", "img",
        "[ATTR]", "src",
        "data", "image", "svg", "xml", "base64", "[BASE64]", "svg",
        "[ATTR]", "width",
        "256",
    ]
    assert tokens.count("img") == 1
    start = tokens.index("img") - 1
    assert start > 0
    limit = start + len(expect)
    assert tokens[start:limit] == expect



import json

from ...util import load_resource


def test_raw_response_unwrapping(pre_tokenizer):
    """Test that DOM snapshots wrapped in CDP browser responses are
    pretokenized identically to non-wrapped DOM snapshots like those
    in https://huggingface.co/datasets/gbenson/webui-dom-snapshots.
    """
    wrapped_snapshot = load_resource("raw-browser-response.json")
    browser_response = json.loads(wrapped_snapshot)
    assert set(browser_response.keys()) == {"id", "result", "sessionId"}
    regular_snapshot = browser_response["result"]
    assert set(regular_snapshot.keys()) == {"documents", "strings"}
    regular_snapshot = json.dumps(regular_snapshot, separators=(",", ":"))
    assert regular_snapshot in wrapped_snapshot
    del browser_response

    regular_tokens = pre_tokenizer.tokenize(regular_snapshot)
    assert regular_tokens.count("[TAG]") == 5
    assert regular_tokens.count("hello") == 2

    wrapped_tokens = pre_tokenizer.tokenize(wrapped_snapshot)
    assert wrapped_tokens == regular_tokens



import pytest


class PreTokenizerTester:
    def __init__(self, pre_tokenizer):
        self.wrapped_pre_tokenizer = pre_tokenizer

    @property
    def pre_tokenize_str(self):
        return self.wrapped_pre_tokenizer.pre_tokenize_str

    def tokenize(self, *args, **kwargs):
        return [
            token
            for token, offsets in self.pre_tokenize_str(*args, **kwargs)
        ]


@pytest.fixture()
def pre_tokenizer(dom_snapshot_pre_tokenizer):
    return PreTokenizerTester(dom_snapshot_pre_tokenizer)
